"""
Utilities for Mop-utils.
"""


class ImageAnalysisInput:
    data: bytes = b''


class TextAnalysisInput:
    text: str = ""


class AnalysisResult:
    harmful_score: float = 0.0
    severity_level: int = 0


class AcsImageResponse:
    """
    All four fields "hate", "self_harm", "sexual" and "violence" are required.
    """
    hate = AnalysisResult()
    self_harm = AnalysisResult()
    sexual = AnalysisResult()
    violence = AnalysisResult()


class AcsTextResponse:
    """
    All four fields "hate", "self_harm", "sexual" and "violence" are required.
    """
    hate = AnalysisResult()
    self_harm = AnalysisResult()
    sexual = AnalysisResult()
    violence = AnalysisResult()


class MopInferenceOutputValidator:
    """
    MOP inference output validator.
    """

    def __init__(self, mop_inference_output):
        self.predicted_labels = mop_inference_output.predicted_labels
        self.confidence_scores = mop_inference_output.confidence_scores
        if not self.predicted_labels or not self.confidence_scores:
            raise ValueError(f"Invalid predicted_labels and confidence_scores. "
                             f"Current value: {self.confidence_scores}, {self.predicted_labels}")

    def _check_predicted_labels(self):
        """
           Validate predicted label name, type, and validate its value's key name, key type and key value.
           @return:
           @rtype:
       """
        if self.predicted_labels:
            for key, value in self.predicted_labels.items():
                if not key or not isinstance(key, str):
                    raise TypeError(f"The label name must be type str and its value must be not empty. "
                                    f"Current value: {key}, type: {type(key)}")

                if not value or not isinstance(value, dict):
                    raise TypeError(f"The label scores must be dict type. Current value: {value}, type: {type(value)}")

                for k, v in value.items():
                    if not k or not isinstance(k, str):
                        raise TypeError(f"The key should be type str and not empty. "
                                        f"Current value: {k}, type: {type(k)}")

                    if v is None or not isinstance(v, int) or v not in (1, 0):
                        raise TypeError(f"The value should be int ( 1 or 0 ): "
                                        f"Current key: {k}, value: {v}, type: {type(v)}")

                if len(value.keys()) == 0:
                    raise ValueError(f"There must be at least one value. ")

    def _check_confidence_scores(self):
        """
            Validate confidence score label name, type, and validate its value's key name, key type, key value.
        """
        if self.confidence_scores:
            for key, value in self.confidence_scores.items():
                if not key or not isinstance(key, str):
                    raise TypeError(f"The label name must be str and not empty. "
                                    f"Current value: {key}, type: {type(key)}")

                if not value or not isinstance(value, dict):
                    raise TypeError(f"The confidence scores must be dict and also not empty. "
                                    f"Current value: {value}, type: {type(value)}")

                for k, v in value.items():
                    if not k or not isinstance(k, str):
                        raise TypeError(f"The confidence score key name must be str and also not empty. "
                                        f"Current value: {k}, type: {type(k)}")

                    if v is None or not isinstance(v, (float, int)):
                        raise TypeError(f"The confidence score value must be float or int and not empty. "
                                        f"Current key: {k}, value: {v}, type: {type(v)}")

                if len(value.keys()) == 0:
                    raise ValueError(f"There must be at least one key in confidence score.")

    def _check_keys_match(self) -> bool:
        """
        Check if predict labels keys match confidence score keys.
        """
        if len(self.predicted_labels.keys()) != len(self.confidence_scores.keys()):
            return False

        if set(self.predicted_labels.keys()) != set(self.confidence_scores.keys()):
            return False

        return True

    def _check_value_keys_match(self) -> bool:
        """
        Check if the keys of predict label value match the keys of confidence score value.
        """
        if not self.predicted_labels and not self.confidence_scores:
            return True

        for key in self.confidence_scores.keys():
            conf_score = self.confidence_scores[key]
            pre_label = self.predicted_labels.get(key, None)
            conf_keys = conf_score.keys()
            pre_keys = pre_label.keys()
            if set(conf_keys) != set(pre_keys):
                return False

        return True

    def validate(self):
        """
        Validate if label name, type, the numbers of label match between confidence score and predicted labels.
        """

        self._check_predicted_labels()
        self._check_confidence_scores()

        if not self._check_keys_match():
            raise ValueError(f"The keys of predicted labels should match the keys of confidence scores. Current: "
                             f"{self.predicted_labels.keys()}, {self.confidence_scores.keys()}")

        if not self._check_value_keys_match():
            raise ValueError(f"The keys of predicted labels values should match the keys of confidence score values. "
                             f"Current: {self.confidence_scores}, {self.predicted_labels}")



import importlib
import json
import os
from typing import Any, Dict, List, Optional
from pathlib import Path
import numpy as np
from add_core_mop.base_model_wrapper import BaseModelWrapper, MopInferenceInput
from add_core_mop.constant import CM_MODEL_WRAPPER_NAME
from pyraisdk.dynbatch import BaseModel, DynamicBatchModel

inference_module = importlib.import_module(CM_MODEL_WRAPPER_NAME)
wrappers = []
for i in inference_module.__dict__.keys():
    if hasattr(inference_module.__dict__.get(i), '__bases__'):
        if BaseModelWrapper in inference_module.__dict__.get(i).__bases__[0].__bases__:
            wrappers.append(getattr(inference_module, i))
        if BaseModelWrapper in inference_module.__dict__.get(i).__bases__:
            wrappers.append(getattr(inference_module, i))

ModelWrapper = [i for i in wrappers if i.__module__ == CM_MODEL_WRAPPER_NAME][0]


class MOPInferenceWrapper:
    def __init__(self, base_model_wrapper: BaseModelWrapper) -> None:
        self.model_wrapper = base_model_wrapper

    def init(self, model_root: str) -> None:
        self.model_wrapper.init(model_root)

    def run(self, item: Dict, triggered_by_mop) -> Dict:
        if not triggered_by_mop:
            model_output = self.model_wrapper.inference(item)
            return [model_output]
        else:
            mop_input = MopInferenceInput().from_dict(item)
            model_input = self.model_wrapper.convert_mop_input_to_model_input(mop_input)
            model_output = self.model_wrapper.inference(model_input)
            mop_output = self.model_wrapper.convert_model_output_to_mop_output(model_output)

            return mop_output.output

    def run_batch(self, items: List[dict], triggered_by_mop: bool, batch_size: Optional[int] = None) -> List[dict]:
        if not triggered_by_mop:
            model_outputs = self.model_wrapper.inference_batch(items)
            return model_outputs
        else:
            model_inputs = [
                self.model_wrapper.convert_mop_input_to_model_input(MopInferenceInput().from_dict(item)) for item
                in
                items]

            model_outputs = self.model_wrapper.inference_batch(model_inputs)

            mop_outputs = [self.model_wrapper.convert_model_output_to_mop_output(model_output).output for
                           model_output in model_outputs]
            return mop_outputs


batch_model: Optional[DynamicBatchModel] = None
base_model_wrapper: BaseModelWrapper = ModelWrapper()
inference_wrappers: Dict[str, MOPInferenceWrapper] = {}
inference_wrapper: MOPInferenceWrapper = MOPInferenceWrapper(base_model_wrapper)
batch_size: Optional[int] = None
triggered_by_mop: bool = False


class WrapModel(BaseModel):
    def predict(self, items: List[Any]) -> List[Any]:
        return inference_wrapper.run_batch(items, triggered_by_mop=triggered_by_mop, batch_size=batch_size)


class NumpyJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.bool_):
            return bool(obj)
        return json.JSONEncoder.default(self, obj)


"""
This is init function.

Parameters:
    model_root - root where the model file exists

Returns:
    None
"""


def check_model_directory(directory):
    directories = [d for d in os.listdir(directory) if os.path.isdir(os.path.join(directory, d))]
    if 'models' in directories:
        return True
    else:
        return False


def mop_init(model_root, dynamic_batch_args: None):
    global batch_model
    global batch_size
    global base_model_wrapper
    p = Path(model_root)

    if check_model_directory(p):
        models_dir = p / 'models'
        index = 0
        for child in models_dir.iterdir():

            model_name = child.stem
            wrapper = MOPInferenceWrapper(ModelWrapper())
            model_each_root = os.path.join(str(child), 'model')

            print(f"test number: {index}, model_each_root: {model_each_root}")
            wrapper.init(model_each_root)
            inference_wrappers[model_name] = wrapper
            index = index + 1
    else:
        print(f"else case： {model_root}")
        inference_wrapper.init(model_root)

    # assign environment variable
    if dynamic_batch_args is not None:
        os.environ["PYRAISDK_MAX_BATCH_SIZE"] = str(dynamic_batch_args.get('max_batch_size'))
        os.environ["PYRAISDK_IDLE_BATCH_SIZE"] = str(dynamic_batch_args.get('idle_batch_size'))
        os.environ["PYRAISDK_MAX_BATCH_INTERVAL"] = str(dynamic_batch_args.get('max_batch_interval'))

        batch_model = DynamicBatchModel(WrapModel())
        batch_size = dynamic_batch_args.get('max_batch_size')


"""
This is run function.

Parameters:
    raw_data - row input data to do inference
    triggered_by_mop - whether the function is triggered by mop
    **kwargs - dynamic parameter

Returns:
    inference result
    {"holder_type":"8fb40941-9eb7-4bb2-8242-b266c25f1251", "data": {"data": ["hello, Sample Post Data!"]}}
"""


def mop_run(raw_data: any, is_mop_triggered: bool = False, **kwargs) -> any:
    global triggered_by_mop
    global inference_wrapper
    triggered_by_mop = is_mop_triggered
    if isinstance(raw_data, dict) and 'holder_type' in raw_data:
        holder_type = raw_data['holder_type']
        inference_wrapper = inference_wrappers[holder_type]
        raw_data = raw_data['data']
        print("mop_run raw_data contains 'holder_type' key")
    else:
        print("mop_run raw_data is not a dictionary or does not contain 'holder_type' key")

    if batch_model is not None:
        print("mop_run batch_model case")
        raw_data = raw_data if isinstance(raw_data, list) else [raw_data]
        inference_result = batch_model.predict(raw_data, timeout=60)
        return inference_result
    if isinstance(raw_data, dict):
        print("mop_run dict case")
        inference_result = inference_wrapper.run(raw_data, triggered_by_mop)
        return inference_result
    if isinstance(raw_data, list):
        print("mop_run list case")
        inference_result = inference_wrapper.run_batch(raw_data, triggered_by_mop=triggered_by_mop)
        return inference_result
    raise Exception("Invalid input data format")


def build_response(inference_result: any) -> any:
    if isinstance(inference_result, dict):
        output_str = json.dumps(inference_result, cls=NumpyJsonEncoder)
        return json.loads(output_str)
    if isinstance(inference_result, list):
        res = [item.__dict__ if hasattr(inference_result, '__dict__') else item for item in inference_result]
        output_str = json.dumps(res, cls=NumpyJsonEncoder)
        return json.loads(output_str)
    if hasattr(inference_result, '__dict__'):
        output_str = json.dumps(inference_result.__dict__, cls=NumpyJsonEncoder)
        return json.loads(output_str)
    return inference_result


def get_model_wrapper():
    return inference_wrapper


if __name__ == "__main__":
    model_root = "D:\code\carnegie-mop\sample\model"
    # mop_init(model_root, None)
    #
    # text_dict = {"text": "354"}
    # res = mop_run(text_dict, True)
    # print(res)
    # text_dict = [{"text": "354"}]
    # res = mop_run(text_dict, True)
    # print(res)

    dynamic_batch = {
        'enable': True,
        'max_batch_size': 12,
        'idle_batch_size': 3,
        'max_batch_interval': 0.2
    }

    mop_init(model_root, dynamic_batch)

    text_dict = {"text": "354"}
    res = mop_run(text_dict, True)
    print(res)



from abc import ABC, abstractmethod
from typing import List, Dict, Optional, Any
from .util import AcsTextResponse, AcsImageResponse, ImageAnalysisInput, TextAnalysisInput, MopInferenceOutputValidator


class MopInferenceInput:
    def __init__(self, text: Optional[str] = None, image: Optional[str] = None,
                 width: Optional[int] = None,
                 height: Optional[int] = None,
                 images: Optional[List] = None) -> None:
        self.__text__ = text
        self.__image__ = image
        self.__width__ = width
        self.__height__ = height
        self.__images__ = images

    def from_dict(self, input_dict: Dict) -> Any:
        self.__text__ = input_dict.get('text', None)
        self.__image__ = input_dict.get('image', None)
        self.__width__ = input_dict.get('width', None)
        self.__height__ = input_dict.get('height', None)
        self.__images__ = input_dict.get('images', None)
        if self.__text__ is None and self.__image__ is None and self.__images__ is None:
            raise ValueError('Either text or image must be provided')
        return self

    def __str__(self) -> str:
        return str(vars(self))

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def text(self) -> str:
        return self.__text__

    @text.setter
    def text(self, text: str) -> None:
        self.__text__ = text

    @property
    def image(self) -> str:
        return self.__image__

    @image.setter
    def image(self, image: str) -> None:
        self.__image__ = image

    @property
    def width(self) -> int:
        return self.__width__

    @width.setter
    def width(self, width: int) -> None:
        self.__width__ = width

    @property
    def height(self) -> int:
        return self.__height__

    @height.setter
    def height(self, height: int) -> None:
        self.__height__ = height

    @property
    def images(self) -> str:
        return self.__images__

    @images.setter
    def images(self, images: str) -> None:
        self.__images__ = images


class MopInferenceOutput:

    def __init__(self, raw_data_dict: dict) -> None:
        """
        Initialize MopInferenceOutput.

        The label names for each taxonomy can be defined based on the specific model.
        In order to be deemed a valid MopInferenceOutput, certain requirements must be met:

        1. The key names 'predicted_labels' and 'confidence_scores' are required, and the label names
        corresponding to the same taxonomy in both must be identical.

        2. When the label type is categorical, there must be at least two label names for each taxonomy.
        Among these labels, only one should have a value of 1, while the rest should be set to 0.

        When dealing with ordinal label type, it's important to note that each taxonomy should consist of a minimum
        of three label names. These labels should be organized in ascending order based on the sequence information
        of the label itself, rather than their position in the dictionary.

        For example:
        If the label type is categorical, the names and values of label for each taxonomy can be defined as follows:
        {
            "predicted_labels": {
                "violence": {
                    "severity-1": 0,
                    "severity-2": 1,
                    "severity-3": 0
                },
                "hate": {
                    "sev-1": 0,
                    "sev-2": 0,
                    "sev-3": 1
                }
            },
            "confidence_scores": {
                "violence": {
                    "severity-1": 0.325,
                    "severity-2": 0.675,
                    "severity-3": 0
                },
                "hate": {
                    "sev-1": 0.125,
                    "sev-2": 0.325,
                    "sev-3": 0.55
                }
            }
        }

        If the label type is ordinal, the names and values of label for each taxonomy can be defined as follows:
        {
            "predicted_labels": {
                "violence": {
                    "severity-1": 1,
                    "severity-2": 1,
                    "severity-3": 0
                },
                "hate": {
                    "sev-1": 1,
                    "sev-2": 0,
                    "sev-3": 0
                }
            },
            "confidence_scores": {
                "violence": {
                    "severity-1": 0.5485,
                    "severity-2": 0.525,
                    "severity-3": 0
                },
                "hate": {
                    "sev-1": 0.26544,
                    "sev-2": 0.225,
                    "sev-3": 0.225
                }
            }
        }
        """
        self.predicted_labels = raw_data_dict.get("predicted_labels", None)
        self.confidence_scores = raw_data_dict.get("confidence_scores", None)
        if not self.__predicted_labels__ or not self.__confidence_scores__:
            raise ValueError(f"MopInferenceOutput: invalid raw data dict: Current value: {raw_data_dict}")

        MopInferenceOutputValidator(self).validate()

    def from_dict(self, raw_data_dict: dict) -> Any:
        """
        Deprecated: This method will be deprecated.
        """
        self.__confidence_scores__ = raw_data_dict['confidence_scores']
        self.__predicted_labels__ = raw_data_dict['predicted_labels']
        MopInferenceOutputValidator(self).validate()
        return self

    def __str__(self) -> str:
        return str(vars(self))

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def confidence_scores(self) -> Dict:
        return self.__confidence_scores__

    @confidence_scores.setter
    def confidence_scores(self, confidence_scores: Dict) -> None:
        self.__confidence_scores__ = confidence_scores

    @property
    def predicted_labels(self) -> Dict:
        return self.__predicted_labels__

    @predicted_labels.setter
    def predicted_labels(self, predicted_labels: Dict) -> None:
        self.__predicted_labels__ = predicted_labels

    @property
    def output(self) -> Dict:
        return {'confidence_scores': self.confidence_scores, 'predicted_labels': self.predicted_labels}


class BaseModelWrapper(ABC):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def init(self, model_root: str, **kwargs):
        raise NotImplementedError("init() method is not implemented")

    @abstractmethod
    def inference(self, item: Dict, **kwargs) -> Dict:
        raise NotImplementedError("inference() method is not implemented")

    @abstractmethod
    def inference_batch(self, item: List[Dict], **kwargs) -> List[Dict]:
        raise NotImplementedError("inference_batch() method is not implemented")

    @abstractmethod
    def convert_mop_input_to_model_input(self, mop_input: MopInferenceInput, **kwargs) -> Dict:
        raise NotImplementedError("convert_mop_input_to_model_input() method is not implemented")

    @abstractmethod
    def convert_model_output_to_mop_output(self, customized_output: Dict, **kwargs) -> MopInferenceOutput:
        raise NotImplementedError("convert_model_output_to_mop_output() method is not implemented")

    def convert_acs_text_request_to_model_inference_input(self, req: TextAnalysisInput) -> object:
        """
        Optional implementation: Convert ACS text request to model inference input.
        But it is mandatory if you want to release the model to AACS service.
        @param req: ACS text request
        @type req: TextAnalysisInput
        @return: Model inference input
        @rtype: object
        """
        pass

    def convert_model_inference_output_to_acs_text_response(self, out: object) -> AcsTextResponse:
        """
        Optional implementation: Convert model inference output to ACS text response.
        But it is mandatory if you want to release the model to AACS service.
        @param out: Model inference output
        @type out: object
        @return: ACS text response
        @rtype: AcsTextResponse
        """
        pass

    def convert_acs_image_request_to_model_inference_input(self, req: ImageAnalysisInput) -> object:
        """
        Optional implementation: Convert ACS image request to model inference input.
        But it is mandatory if you want to release the model to AACS service.
        @param req: ACS image request
        @type req: ImageAnalysisInput
        @return: Model inference input
        @rtype: object
        """
        pass

    def convert_model_inference_output_to_acs_image_response(self, out: object) -> AcsImageResponse:
        """
        Optional implementation: Convert model inference output to ACS image response.
        But it is mandatory if you want to release the model to AACS service.
        @param out: Model inference output
        @type out: object
        @return: ACS image response
        @rtype: AcsImageResponse
        """
        pass
